from typing import Tuple, List
from kubernetes import client
from kubernetes.client import V1Volume, V1VolumeMount


VOLUME_SOURCES = {
    'aws_elastic_block_store': client.V1AWSElasticBlockStoreVolumeSource,
    'azure_disk': client.V1AzureDiskVolumeSource,
    'azure_file': client.V1AzureFileVolumeSource,
    'cephfs': client.V1CephFSVolumeSource,
    'cinder': client.V1CinderVolumeSource,
    'config_map': client.V1ConfigMapVolumeSource,
    'csi': client.V1CSIVolumeSource,
    'downward_api': client.V1DownwardAPIVolumeSource,
    'empty_dir': client.V1EmptyDirVolumeSource,
    'fc': client.V1FCVolumeSource,
    'flex_volume': client.V1FlexVolumeSource,
    'flocker': client.V1FlockerVolumeSource,
    'gce_persistent_disk': client.V1GCEPersistentDiskVolumeSource,
    'git_repo': client.V1GitRepoVolumeSource,
    'glusterfs': client.V1GlusterfsVolumeSource,
    'host_path': client.V1HostPathVolumeSource,
    'iscsi': client.V1ISCSIVolumeSource,
    'nfs': client.V1NFSVolumeSource,
    'persistent_volume_claim': client.V1PersistentVolumeClaimVolumeSource,
    'photon_persistent_disk': client.V1PhotonPersistentDiskVolumeSource,
    'portworx_volume': client.V1PortworxVolumeSource,
    'projected': client.V1ProjectedVolumeSource,
    'quobyte': client.V1QuobyteVolumeSource,
    'rbd': client.V1RBDVolumeSource,
    'scale_io': client.V1ScaleIOVolumeSource,
    'secret': client.V1SecretVolumeSource,
    'storageos': client.V1StorageOSVolumeSource,
    'vsphere_volume': client.V1VsphereVirtualDiskVolumeSource,
}


def create_volumes(task_volumes) -> Tuple[List[V1Volume], List[V1VolumeMount]]:
    index = 0
    mounts = []
    volumes = []
    for target, volume in task_volumes.items():
        index += 1
        name = volume.get('name', f'volume{index}')
        for source_type, VolumeSource in VOLUME_SOURCES.items():
            if source_type not in volume:
                continue

            volume_config = volume[source_type]
            volumes.append(V1Volume(**{
                'name': name,
                source_type: VolumeSource(**volume_config),
            }))
            mounts.append(V1VolumeMount(
                name=name,
                read_only=volume.get('read_only', False),
                mount_path=target,
            ))

    return volumes, mounts



from cowait.tasks import TaskDefinition, RemoteTask
from cowait.engine.cluster import ClusterProvider


class DockerTask(RemoteTask):
    def __init__(
        self,
        cluster: ClusterProvider,
        taskdef: TaskDefinition,
        container
    ):
        super().__init__(
            cluster=cluster,
            taskdef=taskdef,
        )
        self.container = container
        self.ip = taskdef.id  # id should be routable within docker

    def __str__(self):
        return f'DockerTask({self.id}, {self.status}, {self.inputs}, {self.container.id[:10]})'



# flake8: noqa: F401

from .const import *
from .errors import SocketError, AuthError
from .utils import get_local_url, get_remote_url, get_local_ip
from .client import Client
from .server import Server
from .conn import Conn



# flake8: noqa: F401

from .const import *
from .cluster import ClusterProvider
from .docker import DockerProvider
from .kubernetes import KubernetesProvider
from .api import ApiProvider
from .errors import TaskCreationError, ProviderError
from .utils import env_pack, env_unpack


def get_cluster_provider(type, **args):
    if type == 'docker':
        return DockerProvider(args)
    elif type == 'kubernetes':
        return KubernetesProvider(args)
    elif type == 'api':
        return ApiProvider(args)
    else:
        raise ValueError(f'Unknown cluster provider: {type}')



from cowait.tasks import Task, TaskDefinition


class ContainerTask(Task):
    async def run(
        self, name: str, image: str, env: dict = {},
        routes: dict = {}, ports: dict = {}, meta: dict = {},
        cpu: any = 0, memory: any = 0, affinity: str = None,
        **inputs
    ):
        taskdef = TaskDefinition(
            name=name,
            image=image,
            parent=self.id,
            owner=self.taskdef.owner,
            inputs=inputs,
            ports=ports,
            routes=routes,
            cpu=cpu,
            memory=memory,
            affinity=affinity,
            env={
                **self.taskdef.env,
                **env,
            },
            meta={
                **self.taskdef.meta,
                **meta,
            },
        )

        # run it
        task = self.cluster.spawn(taskdef)

        # wait for container to exit
        await self.watch(task)

        # clean up
        self.cluster.destroy(task.id)

    async def watch(self, task):
        logs = self.cluster.logs(task.id)
        for log in logs:
            print(log)



import sys


class Logger(object):
    def __init__(self):
        self.stdout = sys.stdout
        self.stderr = sys.stderr

    async def handle(self, msg: dict) -> None:
        self.log(msg['type'], msg)

    def log(self, type: str, msg: dict) -> None:
        pass

    def print(self, *args, **kwargs) -> None:
        print(*args, file=self.stdout, **kwargs)

    def print_err(self, *args, **kwargs) -> None:
        print(*args, file=self.stderr, **kwargs)



from .const import DEFAULT_REPO, DEFAULT_BASE_IMAGE


def parse_task_image_name(
    full_name: str,
    default_image: str = DEFAULT_BASE_IMAGE
) -> (str, str):
    image = default_image
    name = full_name
    if '/' in full_name:
        s = full_name.rfind('/')
        image = full_name[:s]
        name = full_name[s+1:]
        if '/' not in image:
            # prepend default repo
            image = f'{DEFAULT_REPO}/{image}'

    return image, name



# flake8: noqa: F401

from .type import Type
from .dict import Dict
from .list import List
from .custom import CustomType
from .simple import Any, String, Int, Float, Bool, DateTime, Void
from .mapping import TypeAlias
from .utils import typed_arguments, typed_return, \
    typed_call, typed_async_call, serialize, deserialize, \
    get_parameter_types, get_parameter_defaults, get_return_type, \
    type_from_description

try:
    from .numpy import *
except ModuleNotFoundError:
    pass



from typing import List
from kubernetes.client import V1EnvVar, V1EnvVarSource, V1SecretKeySelector, \
    V1ConfigMapKeySelector, V1ContainerPort
from cowait.tasks import TaskDefinition
from cowait.engine.utils import base_environment


def create_env(cluster, taskdef: TaskDefinition) -> List[V1EnvVar]:
    """Creates a kubernetes environment configuration"""
    env = base_environment(cluster, taskdef)
    return [
        _parse_env(key, value)
        for key, value in env.items()
    ]


def _parse_env(name: str, value) -> V1EnvVar:
    """Parses an environment value into a kubernetes EnvVar"""
    if isinstance(value, dict):
        return V1EnvVar(
            name=str(name),
            value_from=_parse_env_from(value),
        )
    return V1EnvVar(name=str(name), value=str(value))


def _parse_env_from(value_from: dict) -> V1EnvVarSource:
    """Parses an environment value dict into an EnvVarSource"""
    source = value_from.get('source', '').lower()
    if source == 'secret':
        return V1EnvVarSource(
            secret_key_ref=V1SecretKeySelector(
                name=value_from['name'],
                key=value_from['key'],
            ),
        )
    elif source == 'configmap':
        return V1EnvVarSource(
            config_map_key_ref=V1ConfigMapKeySelector(
                name=value_from['name'],
                key=value_from['key'],
            ),
        )
    else:
        raise ValueError(f'Unsupported environment source "{source}"')


def create_ports(ports: dict) -> List[V1ContainerPort]:
    return [
        V1ContainerPort(**_parse_port(port, host_port))
        for port, host_port in ports.items()
    ]


def _parse_port(port, host_port: str = None) -> dict:
    protocol = 'TCP'
    if isinstance(port, str) and '/' in port:
        slash = port.find('/')
        protocol = port[slash+1:]
        port = port[:slash]

    if host_port is None:
        host_port = port

    return {
        'protocol': protocol.upper(),
        'container_port': int(port),
        'host_port': int(host_port),
    }



import json
import gzip
from base64 import b64encode, b64decode
from cowait.tasks import TaskDefinition
from .const import ENV_TASK_CLUSTER, ENV_TASK_DEFINITION, ENV_GZIP_ENABLED


def env_pack(value):
    value = json.dumps(value).encode('utf-8')
    value = gzip.compress(value)
    return b64encode(value).decode('utf-8')


def env_unpack(value):
    value = b64decode(value)
    value = gzip.decompress(value)
    return json.loads(value)


def base_environment(cluster, taskdef: TaskDefinition) -> dict:
    """
    Create a container environment dict from a task definition.

    Arguments:
        taskdef (TaskDefinition): Task definition

    Returns:
        env (dict): Environment variable dict
    """

    return {
        **taskdef.env,
        ENV_GZIP_ENABLED:    '1',
        ENV_TASK_CLUSTER:    env_pack(cluster.serialize()),
        ENV_TASK_DEFINITION: env_pack(taskdef.serialize()),
    }




import inspect
from cowait.tasks.components import TaskManager


def html_repr(target):
    if not inspect.isclass(target):
        raise AttributeError(f'Target must be a class')

    if hasattr(target, '_repr_html_'):
        raise AttributeError(f'{target.__name__} already has a _repr_html_ method')

    def decorate(fn):
        setattr(target, '_repr_html_', fn)
        return fn

    return decorate


@html_repr(TaskManager)
def task_manager_html(self):
    rows = ''.join([task_row_html(t) for t in self.values()])
    if len(rows) == 0:
        rows = f'<tr><td colspan="3">no tasks</td></tr>'

    return (f'<table>'
            f'<tr>'
            f'<th>id</th>'
            f'<th>status</th>'
            f'<th>inputs</th>'
            f'</tr>'
            f'{rows}'
            f'</table>')


def task_row_html(task):
    return (f'<tr>'
            f'<td>{task.id}</td>'
            f'<td>{task.status}</td>'
            f'<td>{task.inputs}</td>'
            f'</tr>')



import asyncio


async def join(tasks: list) -> list:
    """ Waits for a list of tasks to complete, returning a list containing their results. """
    return await asyncio.gather(*tasks)


async def wait(tasks: list, ignore_errors: bool = False) -> None:
    """ Waits for a list of tasks to complete. Returns nothing. """
    await asyncio.gather(*tasks, return_exceptions=ignore_errors)



import sys
import getpass
from cowait.tasks import TaskDefinition, TASK_LOG
from cowait.engine import ProviderError, TaskCreationError
from ..config import Config
from ..context import Context
from ..utils import ExitTrap
from ..logger import Logger
from .build import build as run_build
from .push import push as run_push


def test(
    config: Config,
    cluster_name: str = None,
    mount: bool = True,
):
    logger = TestLogger()
    try:
        context = Context.open(config)
        cluster = context.get_cluster(cluster_name)

        volumes = {}
        if mount and cluster.type == 'docker':
            # when testing in docker, mount the local directory
            # this avoids the problem of having to constantly rebuild in order to test
            print('** Mounting', context.root_path)
            volumes['/var/task'] = {
                'bind': {
                    'src': context.root_path,
                    'mode': 'rw',
                    'inherit': 'same-image',
                },
            }

        # execute the test task within the current image
        task = cluster.spawn(TaskDefinition(
            name='cowait.test',
            image=context.image,
            owner=getpass.getuser(),
            env={
                **context.environment,
                **context.dotenv,
            },
            volumes={
                **context.get('volumes', {}),
                **volumes,
            },
        ))

        def destroy(*args):
            logger.header('interrupt')
            cluster.destroy(task.id)
            sys.exit(1)

        with ExitTrap(destroy):
            # capture & print logs
            logs = cluster.logs(task.id)
            logger.header('task output')
            for msg in logs:
                logger.handle(msg)

        logger.header()

        # grab task result
        passing = cluster.wait(task)
        sys.exit(0 if passing else 1)

    except ProviderError as e:
        logger.print_exception(f'Provider Error: {e}')
        sys.exit(1)

    except TaskCreationError as e:
        logger.print_exception(f'Error creating task: {e}')
        sys.exit(1)


class TestLogger(Logger):
    def handle(self, msg):
        type = msg.get('type', None)
        if type == TASK_LOG:
            print(msg['data'], end='')



import os
import os.path
from dotenv import dotenv_values
from cowait.utils.const import DEFAULT_BASE_IMAGE
from .utils import find_file_in_parents
from .const import CONTEXT_FILE_NAME
from .config import Config


class Context(Config):
    def __init__(self, root_path: str, *, path: str, parent: Config):
        super().__init__(
            path=path,
            parent=parent,
            data={} if not path else None,
        )
        self.root_path = root_path

    @property
    def workdir(self) -> str:
        return self.get('workdir', '.', False)

    @property
    def image(self):
        return self.get('image', DEFAULT_BASE_IMAGE, False)

    @property
    def base(self):
        return self.get('base', DEFAULT_BASE_IMAGE, False)

    @property
    def environment(self):
        return self.get('environment', {}, False)

    @property
    def notebook(self) -> bool:
        return self.get('features.notebook', False, False)

    @property
    def dotenv(self) -> dict:
        return dotenv_values(self.file('.env'))

    def file(self, file_name: str) -> str:
        """
        Find a file within the task context and return its full path
        """
        path = os.path.join(self.root_path, self.workdir, file_name)
        if not os.path.isfile(path):
            return None
        return path

    def file_rel(self, file_name: str) -> str:
        """
        Find a file within the task context and return its relative path
        """
        abs_path = self.file(file_name)
        if not abs_path:
            return None
        return self.relpath(abs_path)

    def relpath(self, context_path: str):
        """
        Returns a path relative to the context root
        """
        return os.path.relpath(context_path, self.root_path)

    def includes(self, path: str) -> bool:
        """
        Checks if the path is included in the context
        """
        return self.root_path in path

    def pack_data(self, data: dict) -> dict:
        return {
            'version': 1,
            'cowait': data
        }

    def unpack_data(self, data: dict) -> dict:
        if data is None:
            return {}

        if 'version' not in data:
            raise RuntimeError('Invalid context file, no version set.')

        version = int(data['version'])
        if version != 1:
            raise RuntimeError('Wrong context version, expected 1')

        return data.get('cowait', {})

    def write(self, path: str = None) -> None:
        if path is None:
            path = os.path.join(self.root_path, CONTEXT_FILE_NAME)
        return super().write(path)

    @staticmethod
    def exists(path: str = None):
        if path is None:
            path = os.getcwd()

        # ensure the provided path is an actual directory
        if not os.path.isdir(path):
            return False

        # find context root by looking for the context definition file
        context_file_path = find_file_in_parents(path, CONTEXT_FILE_NAME)
        return context_file_path is not None

    @staticmethod
    def open(config: Config, path: str = None):
        if path is None:
            path = os.getcwd()

        # ensure the provided path is an actual directory
        if not os.path.isdir(path):
            raise ValueError(f'Invalid context path {path}: Not a directory')

        # find context root by looking for the context definition file
        context_file_path = find_file_in_parents(path, CONTEXT_FILE_NAME)
        if context_file_path is None:
            # use the current directory as the context
            # no local configuration exists
            return Context(
                root_path=os.path.abspath(path),
                parent=config,
                path=None,
            )

        # context root path is the yml folder
        root_path = os.path.abspath(os.path.dirname(context_file_path))
        return Context(
            root_path=root_path,
            path=context_file_path,
            parent=config,
        )



import os
import psutil


class ResourceMonitor():
    def __init__(self):
        self.process = psutil.Process(os.getpid())
        self.cpu_cores = psutil.cpu_count()
        self.total_memory = psutil.virtual_memory().total
        self.process.cpu_percent()  # initial cpu reading

    def stats(self):
        with self.process.oneshot():
            cpu = self.process.cpu_percent()
            mem = self.process.memory_full_info()
            io = self.process.io_counters()
            fds = self.process.num_fds()

        stats = {
            'cpu': [
                round(cpu / 100.0, 3),
                self.cpu_cores,
            ],
            'mem': [
                mem.uss,  # used
                self.total_memory,  # total
            ],
            'io': [
                io.read_chars,  # read
                io.write_chars,  # write
            ],
            'iops': [
                io.read_count,  # read
                io.write_count,  # write
            ],
            'fds': fds,
        }
        return stats



from typing import List
from kubernetes import client
from .router import Router
from ..const import LABEL_TASK_ID


TRAEFIK2_API_GROUP = 'traefik.containo.us'
TRAEFIK2_API_VERSION = 'v1alpha1'
TRAEFIK2_API_NAMESPACE = 'default'
TRAEFIK2_INGRESSROUTE = 'IngressRoute'
TRAEFIK2_INGRESSROUTE_PLURAL = 'ingressroutes'


class Traefik2Router(Router):
    def __init__(self, cluster):
        super().__init__(cluster)
        cluster.on('prepare', self.on_prepare)
        cluster.on('spawn', self.on_spawn)
        cluster.on('kill', self.on_kill)
        self.config = cluster.args.get('traefik2', {})

    @property
    def secure(self) -> bool:
        return self.cert_resolver is not None

    @property
    def middlewares(self) -> List[str]:
        return self.config.get('middlewares', [])

    @property
    def entrypoints(self) -> List[str]:
        return self.config.get('entrypoints', ['web', 'websecure'])

    @property
    def cert_resolver(self) -> str:
        return self.config.get('certresolver', None)

    @property
    def domain(self) -> str:
        domain = self.cluster.domain
        if domain is None:
            raise RuntimeError('No cluster domain configured')
        return domain

    def on_prepare(self, taskdef):
        protocol = 'https' if self.secure else 'http'

        for path, port in taskdef.routes.items():
            if len(path) < 1 and path[0] != '/':
                raise RuntimeError(f'Paths must start with /, got {path}')

            taskdef.routes[path] = {
                'port': int(port),
                'path': path,
                'url': f'{protocol}://{taskdef.id}.{self.domain}{path}',
            }

        return taskdef

    def on_spawn(self, task):
        ports = []
        routes = []

        idx = 0
        for path, route in task.routes.items():
            port = route['port']
            idx += 1

            ports.append(client.V1ServicePort(
                port=int(port),
                target_port=int(port),
            ))

            host = f'{task.id}.{self.cluster.domain}'
            routes.append({
                'kind': 'Rule',
                'match': f'Host(`{host}`) && PathPrefix(`{path}`)',
                'middlewares': self.middlewares,
                'services': [
                    {'name': task.id, 'port': int(port)}
                ],
            })

        if len(routes) == 0:
            return

        print('~~ creating task ingress', path, '-> port', port)

        self.cluster.core.create_namespaced_service(
            namespace=self.cluster.namespace,
            body=client.V1Service(
                metadata=client.V1ObjectMeta(
                    name=task.id,
                    namespace=self.cluster.namespace,
                    labels={
                        LABEL_TASK_ID: task.id,
                    },
                ),
                spec=client.V1ServiceSpec(
                    selector={
                        LABEL_TASK_ID: task.id,
                    },
                    ports=ports,
                ),
            ),
        )

        ingress = {
            'entryPoints': self.entrypoints,
            'routes': routes,
        }
        if self.secure:
            ingress['tls'] = {
                'certResolver': self.cert_resolver,
            }

        self.cluster.custom.create_namespaced_custom_object(
            group=TRAEFIK2_API_GROUP,
            version=TRAEFIK2_API_VERSION,
            plural=TRAEFIK2_INGRESSROUTE_PLURAL,
            namespace=TRAEFIK2_API_NAMESPACE,
            body={
                'apiVersion': f'{TRAEFIK2_API_GROUP}/{TRAEFIK2_API_VERSION}',
                'kind': TRAEFIK2_INGRESSROUTE,
                'metadata': {
                    'name': f'{task.id}',
                    'namespace': self.cluster.namespace,
                    'labels': {
                        LABEL_TASK_ID: task.id,
                        'ingress-for': task.id
                    },
                },
                'spec': ingress,
            },
        )

    def on_kill(self, task_id):
        try:
            self.cluster.core.delete_namespaced_service(
                namespace=self.cluster.namespace,
                name=task_id,
            )
        except client.rest.ApiException as e:
            if e.status != 404:
                print('!! error deleting Kubernetes Service:', e)

        try:
            self.cluster.custom.delete_namespaced_custom_object(
                group=TRAEFIK2_API_GROUP,
                version=TRAEFIK2_API_VERSION,
                plural=TRAEFIK2_INGRESSROUTE_PLURAL,
                namespace=self.cluster.namespace,
                name=task_id,
            )
        except client.rest.ApiException as e:
            if e.status != 404:
                print('!! error deleting IngressRoute:', e)



import os
import os.path
import sys
import docker
import cowait
from .context import Context
from .docker_file import Dockerfile


class BuildError(RuntimeError):
    pass


class TaskImage(object):
    def __init__(self, context):
        self.context = context
        self.image = None

    @property
    def name(self):
        return self.context.image

    def build(self, base: str, requirements: str = None, buildargs: dict = {}, quiet: bool = False) -> None:
        """ Build task image """

        # create temporary dockerfile
        df = Dockerfile(base)

        # extra features
        if self.context.notebook:
            df.run('pip install jupyterlab --no-cache-dir')
            df.label('cowait.feature.notebook', 'true')

        # install task-specific requirements
        if requirements:
            df.copy(f'./{requirements}', './requirements.txt')
            df.run('pip install -r ./requirements.txt --no-cache-dir')

        # copy source code
        df.copy('.', '.')

        workdir = self.context.workdir
        if workdir != '.':
            df.workdir(os.path.join('/var/task', workdir))

        df.label('cowait.version', cowait.version)
        self.image = TaskImage.build_image(
            dockerfile=str(df),
            path=self.context.root_path,
            tag=f'{self.name}:latest',
            buildargs=buildargs,
            quiet=quiet,
        )

        # tag the image
        self.image.tag(
            repository=self.name,
            tag='latest',
        )

    def push(self):
        """
        Push context image to a remote registry.
        """
        client = docker.from_env()
        logs = client.images.push(
            repository=self.name,
            tag='latest',
            stream=True,
            decode=True,
        )
        return logs

    @staticmethod
    def open(context: Context = None):
        # automatically create context
        if context is None:
            context = Context.open()

        return TaskImage(
            context=context,
        )

    @staticmethod
    def get(name_or_id):
        client = docker.from_env()
        return client.images.get(name_or_id)

    @staticmethod
    def build_image(quiet: bool, **kwargs):
        client = docker.from_env()
        logs = client.api.build(decode=True, rm=True, **kwargs)

        image_hash = None
        for log in logs:
            if 'stream' in log:
                if not quiet:
                    print(log['stream'], end='', flush=True)
            elif 'aux' in log and 'ID' in log['aux']:
                image_hash = log['aux']['ID']
            elif 'errorDetail' in log:
                raise BuildError(log['errorDetail']['message'])
            else:
                if not quiet:
                    print(log)

        if image_hash is not None:
            return TaskImage.get(image_hash)

    @staticmethod
    def pull(name: str, tag: str = 'latest') -> None:
        # remote images always contain a slash.
        if '/' not in name:
            return

        try:
            # if the image already exists, just return it
            # todo: add a policy setting for this, similar to kubernetes
            return TaskImage.get(f'{name}:{tag}')
        except docker.errors.ImageNotFound:
            # we expect this error. pull the image
            pass

        client = docker.from_env()
        logs = client.api.pull(repository=name, tag=tag, stream=True, decode=True)
        sys.stdout.write('   pulling image...')
        sys.stdout.flush()

        progress = {}
        for update in logs:
            if 'id' in update and 'progressDetail' in update:
                id = update['id']
                progress[id] = update['progressDetail']
            if 'errorDetail' in update:
                print('\rError:', update['error'])
                return

            current = 0
            total = 0
            for detail in progress.values():
                current += detail.get('current', 0)
                total += detail.get('total', 0)

            if total > 0:
                pct = 100 * min(current / total, 1.0)
                sys.stdout.write(f'\r   pulling image... {pct:0.2f}%  ')
                sys.stdout.flush()

        sys.stdout.write('\r   pulling image... done   \n')
        sys.stdout.flush()



# flake8: noqa: F401

from .test_task import PytestTask
from .provider import get_test_provider, spawn_test_task, capture_task_events
from .event_log import EventLog
from .async_mock import AsyncMock



import os
import os.path
from ..const import CONTEXT_FILE_NAME
from ..config import Config


def new_context(
    config: Config,
    name: str,
    image: str,
    base: str,
):
    path = os.getcwd()

    # create directory
    current_dir = os.path.basename(path)
    if current_dir != name and name is not None:
        path = os.path.join(path, name.lower())
        os.mkdir(path)
        print('Created context folder', path)

    context = {}

    # image name
    if image is not None:
        print('Image:', image)
        context['image'] = image

    # base image name
    if base is not None:
        print('Base image:', base)
        context['base'] = base
    
    context = Config(data={
        "version": 1,
        "cowait": context
    }, parent=config)

    context_file = os.path.join(path, CONTEXT_FILE_NAME)
    if os.path.isfile(context_file):
        print('Error: Context file', context_file, 'already exists')
        return

    context.write(context_file)
    print('Created new context definition', context_file)



from kubernetes import client
from cowait.engine.const import LABEL_TASK_ID


affinity_schema = {
    'title': 'TaskAffinity',
    'type': 'object',

    'properties': {
        'mode': {
            'type': 'string',
            'enum': ['stack', 'spread'],
        },
        'required': {
            'type': 'bool',
        },
        'weight': {
            'type': 'number',
            'minimum': 0,
            'maximum': 100,
        },
        'label': {
            'type': 'string',
        },
        'namespaces': {
            'type': 'array',
            'items': {
                'type': 'string',
            },
        },
        'selectors': {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'key': {
                        'type': 'string',
                    },
                    'operator': {
                        'type': 'string',
                        'enum': ['In', 'NotIn', 'Exists', 'DoesNotExist'],
                    },
                    'values': {
                        'type': 'array',
                        'items': {'type': 'string'},
                    },
                },
                'required': ['key', 'operator'],
            },
        },
    },
    'required': ['mode'],
}


def parse_affinity_item(affinity):
    return {
        'mode': affinity.get('mode', 'stack'),
        'required': affinity.get('required', False),
        'label': affinity.get('label', 'kubernetes.io/hostname'),
        'weight': affinity.get('weight', 1),
        'namespaces': affinity.get('namespaces', None),
        'selectors': affinity.get('selectors', [
            {'key': LABEL_TASK_ID, 'operator': 'Exists'},
        ]),
    }


def create_affinity_selector(selector):
    return client.V1LabelSelectorRequirement(
        key=selector.get('key'),
        operator=selector.get('operator'),
        values=selector.get('values', []),
    )


def create_affinity_term(item):
    return client.V1WeightedPodAffinityTerm(
        weight=item['weight'],
        pod_affinity_term=client.V1PodAffinityTerm(
            topology_key=item['label'],
            namespaces=item['namespaces'],
            label_selector=client.V1LabelSelector(
                match_expressions=[create_affinity_selector(s) for s in item['selectors']],
            ),
        )
    )


def create_affinity(affinity):
    affinities = []
    if affinity is None:
        return None
    elif isinstance(affinity, str):
        affinities = [{'mode': affinity}]
    elif isinstance(affinity, dict):
        affinities = [affinity]
    elif isinstance(affinity, list):
        pass
    else:
        raise ValueError('Illegal affinity definition')

    # fill with defaults
    affinities = [parse_affinity_item(item) for item in affinities]

    # sort into required/preferred, affinity/anti-affinity
    stack_req, stack_pref = [], []
    spread_req, spread_pref = [], []
    for item in affinities:
        term = create_affinity_term(item)
        if item['mode'] == 'stack':
            if item['required']:
                stack_req.append(term)
            else:
                stack_pref.append(term)
        elif item['mode'] == 'spread':
            if item['required']:
                spread_req.append(term)
            else:
                spread_pref.append(term)

    return client.V1Affinity(
        pod_affinity=client.V1PodAffinity(
            required_during_scheduling_ignored_during_execution=stack_req,
            preferred_during_scheduling_ignored_during_execution=stack_pref,
        ) if len(stack_req) + len(stack_pref) > 0 else None, 
        pod_anti_affinity=client.V1PodAntiAffinity(
            required_during_scheduling_ignored_during_execution=spread_req,
            preferred_during_scheduling_ignored_during_execution=spread_pref,
        ) if len(spread_req) + len(spread_pref) > 0 else None,
    )




# flake8: noqa: F401

from .router import Router
from .local_port_router import LocalPortRouter
from .traefik_router import TraefikRouter
from .traefik2_router import Traefik2Router


def create_router(cluster, kind: str = 'none'):
    if kind == 'none':
        return Router(cluster)
    if kind == 'local':
        return LocalPortRouter(cluster)
    if kind == 'traefik':
        return TraefikRouter(cluster)
    if kind == 'traefik2':
        return Traefik2Router(cluster)
    
    raise RuntimeError(f'Unknown router type {kind}')



import inspect
from typing import Any
from .task import Task


def task(func):
    """ Wraps a task function in a Task class """

    if func.__name__[0].lower() == func.__name__[0]:
        raise NameError(
            f'Task names must start with an uppercase character, '
            f'found {func.__name__}')

    class FuncTask(Task):
        async def run(self, **inputs):
            if inspect.iscoroutinefunction(func):
                return await func(**inputs)
            else:
                return func(**inputs)

    # copy name & module
    FuncTask.__name__ = func.__name__
    FuncTask.__module__ = func.__module__
    FuncTask.__wraps__ = func

    return FuncTask


def spawn(*args, **kwargs):
    """Spawns subtasks from functional tasks. API is identical to Task.spawn()"""
    task = Task.get_current()
    if task is None:
        raise RuntimeError('No running task found')
    return task.spawn(*args, **kwargs)


def input(name: str, default: Any = None) -> Any:
    """Returns the value of an input to the currently running task"""
    task = Task.get_current()
    if task is None:
        raise RuntimeError('No running task found')

    if default is None and name not in task.taskdef.inputs:
        raise ValueError(f'Input {name} has no default value')

    return task.taskdef.inputs.get(name, default)


def exit(result: Any = None) -> None:
    """Exits the currently running task with the provided result"""
    task = Task.get_current()
    if task is None:
        raise RuntimeError('No running task found')
    task.exit(result)



import asyncio
from cowait.tasks import TaskDefinition, RemoteTask
from cowait.engine.cluster import ClusterProvider
from cowait.engine.errors import TaskCreationError
from .pod import pod_is_ready
from .errors import PodUnschedulableError, PodTerminatedError, ImagePullError


class KubernetesTask(RemoteTask):
    def __init__(
        self,
        cluster: ClusterProvider,
        taskdef: TaskDefinition,
        pod,
    ):
        super().__init__(
            cluster=cluster,
            taskdef=taskdef,
        )
        self.pod = pod
        self.ip = self.pod.status.pod_ip

    def __str__(self):
        return f'KubernetesTask({self.id}, {self.status}, {self.inputs})'

    async def wait_for_scheduling(self):
        poll_interval = 0
        while True:
            await asyncio.sleep(poll_interval)
            pod = self.cluster.get_task_pod(self.id)

            try:
                if pod_is_ready(pod):
                    break
                else:
                    poll_interval = 1

            except PodUnschedulableError as e:
                poll_interval = 10
                print('warning: task', self.id, 'is unschedulable:', str(e))

            except PodTerminatedError:
                raise TaskCreationError('Task terminated') from None
            
            except ImagePullError:
                self.cluster.kill(self.id)
                raise TaskCreationError('Image pull failed')




from aiohttp import WSCloseCode
from datetime import datetime
from .rpc_client import RpcClient


class Conn(object):
    def __init__(self, ws, remote):
        self.ws = ws
        self.id = None
        self.rpc = RpcClient(ws)
        self.remote = remote

    async def send(self, msg: dict) -> None:
        msg['ts'] = datetime.now().isoformat()
        if 'type' not in msg:
            raise Exception('Messages must have a type field')

        await self.ws.send_json(msg)

    async def close(self, message=''):
        self.rpc.cancel_all()
        await self.ws.close(
            code=WSCloseCode.GOING_AWAY,
            message=message
        )



import yaml
import click
import cowait.cli.commands
from cowait.cli import CliError
from .utils import parse_input_list


@click.group(help='start notebook', invoke_without_command=True)
@click.option('-c', '--cluster',
              default=None,
              type=str,
              help='cluster name')
@click.option('-i', '--image',
              type=str,
              default=None,
              help='remote image')
@click.pass_context
def notebook(ctx, cluster, image):
    if ctx.invoked_subcommand is None:
        cowait.cli.notebook(ctx.obj, image, cluster_name=cluster)


@notebook.command(help='run notebook as a task')
@click.argument('path', type=str)
@click.option('--image',
              default=None,
              type=str,
              help='image name')
@click.option('-c', '--cluster',
              default=None,
              type=str,
              help='cluster name')
@click.option('-n', '--name',
              type=str,
              default=None,
              help='specific task name')
@click.option('-i', '--input',
              type=str,
              multiple=True,
              help='specify task input')
@click.option('-e', '--env',
              type=str,
              multiple=True,
              help='define enviornment variable')
@click.option('-b', '--build',
              type=bool, is_flag=True,
              help='build and push first',
              default=False)
@click.option('-d', '--detach',
              type=bool, is_flag=True,
              help='run in detached mode',
              default=False)
@click.option('-f', '--json', '--yml', '--yaml', 'file',
              help='yaml/json file with inputs',
              type=str,
              default=None)
@click.option('-q', '--quiet',
              type=bool, is_flag=True,
              help='no output except result',
              default=False)
@click.pass_context
def run(
    ctx, path: str, image: str, cluster: str, name: str,
    input, env, build: bool, detach: bool, file: str, quiet: bool
):
    file_inputs = {}
    if file is not None:
        try:
            with open(file, 'r') as f:
                file_inputs = yaml.load(f, Loader=yaml.FullLoader)
        except yaml.parser.ParserError as e:
            raise CliError(f'Error in {file}: {e}')

    if not isinstance(file_inputs, dict):
        raise CliError('Error: Expected input file to contain a dictionary')

    cowait.cli.run_notebook(
        ctx.obj,
        path,
        image=image,
        cluster=cluster,
        name=name,
        inputs={
            **file_inputs,
            **parse_input_list(input),
        },
        env=parse_input_list(env),
        build=build,
        detach=detach,
        quiet=quiet)
